import morpion, random
from tqdm import tqdm
import Joueur_QL as JQL
import Joueurs
import numpy as np
import matplotlib.pyplot as plt

EPISODES_ENTRAINEMENT = 5000
EPISODES_STATS = 1000
NOM_IA = "IA_QL"

def obs2state(obs):
    array = np.array(obs)
    state = sum([n * 3**i for i,n in enumerate((array[1] + 2*array[2]).reshape(1,9)[0])])
    return state

# Environnement

env = morpion.Morpion()  # La classe Morpion vient du fichier morpion.py importé à la 1ere ligne

joueur_Aleatoire = Joueurs.Joueur_Aleatoire() # Les classes des différents joueurs proviennent du fichier Joueurs.py importés en début de code
joueur_Simple = Joueurs.Joueur_Simple()
joueur_Avance = Joueurs.Joueur_Avance()
# joueur_Intermediaire = Joueurs.Joueur_Intermediaire()

liste_adversaires_entrainement = [joueur_Aleatoire, joueur_Simple, joueur_Avance]

liste_adversaires_tests = [joueur_Aleatoire, joueur_Simple, joueur_Avance]

joueur = JQL.Joueur(nb_states=3**9, nb_actions=9, file=NOM_IA) # Entrainements consécutifs

for adversaire_entrainement in liste_adversaires_entrainement:
    # Joueur Q-learning

    #joueur = JQL.Joueur(nb_states=3**9, nb_actions=9)

    # Entrainement

    print(f"Début de l'entraînement contre {adversaire_entrainement}...")

    for episode in tqdm(range(EPISODES_ENTRAINEMENT)):

        # Initialisation du jeu
        obs = env.reset()
        env.current_player = random.choice([0,1])
        info = env.info()
        done = False

        # Sauvegarde des différents états du jeu rencontrés et choix associés
        exp_replay_data = [[], []]

        # Facteur d'aléatoire dans le choix de l'IA (décroit au fil des parties)
        coeff_random = 1 / (1+episode)

        while not done:

            if info['current_player'] == 0: # L'IA joue
                state = obs2state(obs[0])
                action = joueur.jouer(state, info['legal_actions'], coeff_random=coeff_random)
            else:
                state = obs2state(obs[1]) # On retient l'état du jeu du point de vue de l'IA (comme si elle jouait)
                action = adversaire_entrainement.jouer(info)

            obs, reward, done, info = env.step(action) # On joue le coup

            # On calcule le nouvel état, et on enregistre tout pour la récompense en fin de partie
            new_state = obs2state(obs[1-info['current_player']])
            exp_replay_data[1-info['current_player']].append((state, action, new_state))

        if reward[0] != 0 or True:
            joueur.entrainer(exp_replay_data[0], reward[0], learning_rate=0.9) # Coups du joueur
            joueur.entrainer(exp_replay_data[1], reward[1], learning_rate=0.9) # Coups de l'adversaire

    print("Entraînement terminé !")
    print("Lancement des tests...")

    liste_scores = np.zeros([len(liste_adversaires_tests), 3])
    for id, adversaire_tests in enumerate(liste_adversaires_tests):

        scores = [0,0,0]

        print(f"Début des parties contre {adversaire_tests}...")

        for _ in tqdm(range(EPISODES_STATS)):
            obs = env.reset()
            info = env.info()
            done = False

            info['current_player'] = random.choice([0,1])

            while not done:

                if info['current_player'] == 0:
                    state = obs2state(obs[0])
                    action = joueur.jouer(state, info['legal_actions'])
                else:
                    action = adversaire_tests.jouer(info)

                obs, reward, done, info = env.step(action)

            if reward[0] > 0:
                scores[0] += 1
            elif reward[0] < 0:
                scores[1] += 1
            else:
                scores[2] += 1

        mv = round(scores[0] / EPISODES_STATS * 100, 4)
        md = round(scores[1] / EPISODES_STATS * 100, 4)
        mn = round(scores[2] / EPISODES_STATS * 100, 4)

        liste_scores[id,:] = [mv,md,mn]

    X = np.arange(len(liste_adversaires_tests))
    width = 0.2

    fig, ax = plt.subplots()
    ax.bar(X, liste_scores[:,0], width, color='g', label="Victoires")
    ax.bar(X + width, liste_scores[:,1], width, color='r', label="Défaites")
    ax.bar(X + 2*width, liste_scores[:,2], width, color='b', label="Nuls")
    ax.set_ylabel("Scores")
    ax.set_title(f"Entrainement contre {adversaire_entrainement}")
    ax.set_xticks(X + width)
    ax.set_xticklabels(liste_adversaires_tests)
    plt.ylim((0,100))

    ax.legend()
    plt.show()

joueur.save(NOM_IA)