"""

Joueur avec IA de Q-Learning

"""

import pickle
import numpy as np
from math import inf

class Joueur:
    def __init__(self, nb_states, nb_actions, file=None):
        self.__Q = np.zeros([nb_states, nb_actions])

        self.__nb_actions = nb_actions

        self.__file = file

        if file != None:
            self.load()

    def jouer(self, state, allowed_actions, coeff_random=0):
        Q2 = (self.__Q[state,:] + np.random.randn(1, self.__nb_actions) * coeff_random)[0]
        return np.argmax([Q2[a] if a in allowed_actions else -inf for a in range(self.__nb_actions)])

    def load(self):
        try:
            with open(self.__file, 'rb') as f:
                unpickler = pickle.Unpickler(f)
                self.__Q = unpickler.load()
                print("IA Chargée")
        except:
            print("Chargement impossible")

    def recompenser(self, state, action, reward, new_state, learning_rate=0.85, actualisation_factor=0.99):
        self.__Q[state,action] = self.__Q[state,action] + learning_rate * (reward + actualisation_factor * np.max(self.__Q[new_state,:]) - self.__Q[state,action])

    def entrainer(self, data, reward, learning_rate=0.85, actualisation_factor=0.99):
        for d in data[::-1]:
            state, action, new_state = d
            self.recompenser(state, action, reward, new_state, learning_rate=learning_rate, actualisation_factor=actualisation_factor)

    def obs2state(self, obs):
        array = np.array(obs)
        state = sum([n * 3**i for i,n in enumerate((array[1] + 2*array[2]).reshape(1,9)[0])])
        return state

    def save(self, file=None):
        self.__file = file if file != None else self.__file
        with open(self.__file, 'wb') as f:
            pickler = pickle.Pickler(f)
            pickler.dump(self.__Q)

    def Q_table(self):
        return np.array(self.__Q)