import numpy as np

class Morpion:
    def __init__(self, markerJ1 = "X", markerJ2 = 'O'):
        self.__grille = None
        self.__cases_vides = None
        self.__grille_joueurs = None
        self.__grille_points = None # Vérifier rapidement un alignement : J1 = 1, J2 = -1
        self.__done = False
        
        self.current_player = 0
        self.__markers = [markerJ1, markerJ2]

        self.__case2point = {c:(c//3,c%3) for c in range(9)}

        self.__info = {}

        self.reset()
    
    def alignement(self, case): # On passe la case  pour vérifier seulement le nécessaire
        ligne, colonne = self.__case2point[case]

        # Ligne
        if abs(sum(self.__grille_points[ligne,:])) == 3:
            return True
        
        # Colonne
        if abs(sum(self.__grille_points[:,colonne])) == 3:
            return True

        # Diagonale
        if (ligne + colonne) % 2 == 0:
            if (ligne + colonne) % 4 == 0: # Première diagonale
                return abs(sum([self.__grille_points[n,n] for n in [0,1,2]])) == 3
            elif ligne == 1: # Deux diagonales
                return abs(sum([self.__grille_points[n,n] for n in [0,1,2]])) == 3 or abs(sum([self.__grille_points[n,2-n] for n in [0,1,2]])) == 3
            else: # Deuxième diagonale
                return abs(sum([self.__grille_points[n,2-n] for n in [0,1,2]])) == 3
        
        return False

    def info(self):
        # Actions gagnantes (créent un alignement) et défensives (empêchent un alignement)
        self.__info['win_actions'] = []
        self.__info['def_actions'] = []
        
        for case in self.__info['legal_actions']:
            ligne, colonne = self.__case2point[case]

            # Action gagnante
            self.__grille_points[ligne,colonne] = 1 - 2*self.current_player
            if self.alignement(case):
                self.__info['win_actions'].append(case)
            
            # Action défensive
            self.__grille_points[ligne,colonne] = 2*self.current_player-1
            if self.alignement(case):
                self.__info['def_actions'].append(case)
            
            # Retour à la normale
            self.__grille_points[ligne,colonne] = 0

        return self.__info

    def obs(self):
        return np.array([[self.__cases_vides, self.__grille_joueurs[0], self.__grille_joueurs[1]], [self.__cases_vides, self.__grille_joueurs[1], self.__grille_joueurs[0]]])
    
    def render(self):
        for ligne in self.__grille.reshape(3,3):
            print(" ".join([c if c != None else "-" for c in ligne]))

    def reset(self):
        self.__grille = np.full(9, None)
        self.__cases_vides = np.full(9, 1)
        self.__grille_joueurs = np.full([2,9], 0)
        self.__grille_points = np.full([3,3], 0)
        self.__done = False
        self.__info['legal_actions'] = [i for i in range(9)]
        self.current_player = 0
        self.__info['current_player'] = self.current_player
        return self.obs()
    
    def step(self, case, verbose=False):
        reward = [0,0]
        joueur_suivant = 1 - self.current_player

        if case in [0,1,2,3,4,5,6,7,8]:
            if self.__grille[case] == None:
                self.__grille[case] = self.__markers[self.current_player]
                self.__cases_vides[case] = 0
                self.__grille_joueurs[self.current_player,case] = 1
                self.__grille_points[self.__case2point[case]] = 1 - 2*self.current_player # J1 : 1, J2 : -1
                self.__info['legal_actions'].remove(case)

                if self.alignement(case):
                    self.__done = True
                    reward[self.current_player] = 1
                    reward[joueur_suivant] = -1
                    if verbose:
                        print("Alignement !")
            else: # Case déjà jouée
                self.__done = True
                reward[self.current_player] = -1
                reward[1-self.current_player] = 1
                if verbose:
                    print("Case déjà jouée !")
        else: # Case non existante
            self.__done = True
            reward[self.current_player] = -1
            reward[1-self.current_player] = 1
        if len(self.__info['legal_actions']) == 0 and not self.__done:
            self.__done = True
            if verbose:
                print("Match nul !")

        self.current_player = joueur_suivant
        self.__info['current_player'] = self.current_player
        
        return self.obs(), reward, self.__done, self.info()
